/*
 * Project: S.Ha.R.K.
 *
 * Coordinators: 
 *   Giorgio Buttazzo    <giorgio@sssup.it>
 *   Paolo Gai           <pj@gandalf.sssup.it>
 *
 * Authors     : 
 *   Giacomo Guidi	 <giacomo@gandalf.sssup.it>
 *   (see the web pages for full authors list)
 *
 * ReTiS Lab (Scuola Superiore S.Anna - Pisa - Italy)
 *
 * http://www.sssup.it
 * http://retis.sssup.it
 * http://shark.sssup.it
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define PI_MUTEX

#include "kernel/kern.h"
#include "modules/intdrive.h"
#include "modules/edf.h"
#include "modules/hardcbs.h"
#include "modules/rr.h"
#include "modules/dummy.h"

#include "modules/sem.h"
#include "modules/hartport.h"
#include "modules/cabs.h"

#include "modules/pi.h"
#include "modules/nop.h"

#include "bca.h"

#include <drivers/shark_linuxc26.h>
#include <drivers/shark_pci26.h>
#include <drivers/shark_input26.h>
#include <drivers/shark_keyb26.h>
#include <drivers/shark_fb26.h>

#define FRAME_BUFFER_DEVICE 0

/*+ sysyem tick in us +*/
#define TICK 0

/*+ RR tick in us +*/
#define RRTICK 2000

/*+ Interrupt Server +*/
#define INTDRIVE_Q 1000
#define INTDRIVE_T 10000
#define INTDRIVE_FLAG 0

void call_shutdown_task(void *arg);
int device_drivers_init();
int device_drivers_close();
void set_shutdown_task();
TASK shutdown_task_body(void *arg);

PID shutdown_task_PID = -1;

TIME __kernel_register_levels__(void *arg)
{
  struct multiboot_info *mb = (struct multiboot_info *)arg;

  INTDRIVE_register_level(INTDRIVE_Q,INTDRIVE_T,INTDRIVE_FLAG);
  EDF_register_level(EDF_ENABLE_ALL);
  HCBS_register_level(HCBS_ENABLE_ALL, 1);
  RR_register_level(RRTICK, RR_MAIN_YES, mb);
  dummy_register_level();

  SEM_register_module();
  CABS_register_module();

  PI_register_module();
  NOP_register_module();

  kern_init_bca();

  return TICK;
}

TASK __init__(void *arg)
{
  struct multiboot_info *mb = (struct multiboot_info *)arg;

  HARTPORT_init();

  /* Create the shutdown task. It will be activated at RUNLEVEL
     SHUTDOWN */
  set_shutdown_task();

  /* Init the drivers */
  device_drivers_init();

  /* Set the shutdown task activation */
  sys_atrunlevel(call_shutdown_task, NULL, RUNLEVEL_SHUTDOWN);

  __call_main__(mb);

  return (void *)0;
}

#ifdef PI_MUTEX
int app_mutex_init(mutex_t *m)
{
  PI_mutexattr_t attr;
                                                                                                                             
  PI_mutexattr_default(attr);
                                                                                                                             
  return mutex_init(m, &attr);
}
#else
int app_mutex_init(mutex_t *m)
{
  NOP_mutexattr_t attr;
                                                                                                                             
  NOP_mutexattr_default(attr);
                                                                                                                             
  return mutex_init(m, &attr);
}
#endif

void set_shutdown_task() {

  /* WARNING: the shutdown task is a background thread. It cannot execute
	      if the system is overloaded */
  NRT_TASK_MODEL nrt;

  nrt_task_default_model(nrt);
  nrt_task_def_system(nrt);

  shutdown_task_PID = task_create("Shutdown Task",shutdown_task_body,&nrt,NULL);
  if (shutdown_task_PID == NIL) {
    sys_shutdown_message("Error: Cannot create shutdown task\n");
    sys_end();
  }

}

int device_drivers_init() {

  int res;
  KEYB_PARMS kparms = BASE_KEYB;
                                                           
  LINUXC26_register_module();

  PCI26_init();

  INPUT26_init();

  keyb_def_ctrlC(kparms, NULL);

  KEYB26_init(&kparms);

  FB26_init();
                                                                             
  res = FB26_open(FRAME_BUFFER_DEVICE);
  if (res) {
    cprintf("Error: Cannot open graphical mode\n");
    KEYB26_close();
    INPUT26_close();
    sys_end();
  }	                                                                                      
        
  FB26_use_grx(FRAME_BUFFER_DEVICE);
                                                        
  FB26_setmode(FRAME_BUFFER_DEVICE,"800x600-16");
	
  return 0;

}

int device_drivers_close() {
                                                                                                                             
  FB26_close(FRAME_BUFFER_DEVICE);
                                                                                                                             
  KEYB26_close();
        
  INPUT26_close();
                                                                                                                             
  return 0;
                                                                                                                             
}

#define SHUTDOWN_TIMEOUT_SEC 3

void call_shutdown_task(void *arg) 
{
  struct timespec t;

  sys_gettime(&t);
  t.tv_sec += SHUTDOWN_TIMEOUT_SEC;

  /* Emergency timeout to exit from RUNLEVEL_SHUTDOWN */
  kern_event_post(&t,(void *)((void *)sys_abort_shutdown),(void *)0);

  task_activate(shutdown_task_PID);
}

TASK shutdown_task_body(void *arg) {

  device_drivers_close();

  sys_shutdown_message("-- S.Ha.R.K. Closed --\n");

  sys_abort_shutdown(0);

  return NULL;

}
